@@ -2,18 +2,21 @@
 /*
  * %CopyrightBegin%
  *
- * Copyright Ericsson AB 1998-2013. All Rights Reserved.
+ * SPDX-License-Identifier: Apache-2.0
  *
- * The contents of this file are subject to the Erlang Public License,
- * Version 1.1, (the "License"); you may not use this file except in
- * compliance with the License. You should have received a copy of the
- * Erlang Public License along with this software. If not, it can be
- * retrieved online at http://www.erlang.org/.
+ * Copyright Ericsson AB 1998-2025. All Rights Reserved.
  *
- * Software distributed under the License is distributed on an "AS IS"
- * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
- * the License for the specific language governing rights and limitations
- * under the License.
+ * Licensed under the Apache License, Version 2.0 (the "License");
+ * you may not use this file except in compliance with the License.
+ * You may obtain a copy of the License at
+ *
+ *     http://www.apache.org/licenses/LICENSE-2.0
+ *
+ * Unless required by applicable law or agreed to in writing, software
+ * distributed under the License is distributed on an "AS IS" BASIS,
+ * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
+ * See the License for the specific language governing permissions and
+ * limitations under the License.
  *
  * %CopyrightEnd%
  */
@@ -25,13 +28,14 @@
 #include "epmd_int.h"
 #include "erl_printf.h" /* erts_snprintf */
 
-#ifndef INADDR_NONE
-#  define INADDR_NONE 0xffffffff
+#ifdef __clang_analyzer__
+   /* CodeChecker does not seem to understand inline asm in FD_ZERO */
+#  undef FD_ZERO
+#  define FD_ZERO(FD_SET_PTR) memset(FD_SET_PTR, 0, sizeof(fd_set))
 #endif
 
-#if defined(__OSE__)
-#  include "sys/ioctl.h"
-#  define sleep(x) delay(x*1000)
+#ifndef INADDR_NONE
+#  define INADDR_NONE 0xffffffff
 #endif
 
 /*
@@ -42,7 +46,7 @@
  *  register their names with this server.
  *
  *  To be accessible to all Erlang nodes this server listens to a well
- *  known port EPMD_PORT_NO (curently port 4369) where requests
+ *  known port EPMD_PORT_NO (currently port 4369) where requests
  *  for connections can be sent.
  *
  *  To keep track of when registered Erlang nodes are terminated this
@@ -63,7 +67,7 @@
 
 /* We use separate data structures for node names and connections
    so that a request will not use a slot with a name that we
-   want to resuse later incrementing the "creation" */
+   want to reuse later incrementing the "creation" */
 
 
 /* forward declarations */
@@ -75,6 +79,7 @@ static time_t current_time(EpmdVars*);
 
 static Connection *conn_init(EpmdVars*);
 static int conn_open(EpmdVars*,int);
+static int conn_local_peer_check(EpmdVars*, int);
 static int conn_close_fd(EpmdVars*,int);
 
 static void node_init(EpmdVars*);
@@ -201,14 +206,52 @@ static EPMD_INLINE void select_fd_set(EpmdVars* g, int fd)
     }
 }
 
+#ifdef HAVE_SOCKLEN_T
+static const char *epmd_ntop(struct sockaddr_storage *sa, char *buff, socklen_t len) {
+#else
+static const char *epmd_ntop(struct sockaddr_storage *sa, char *buff, size_t len) {
+#endif
+    /* Save errno so that it is not changed by inet_ntop */
+    int myerrno = errno;
+    const char *res;
+#if defined(EPMD6)
+    if (sa->ss_family == AF_INET6) {
+        struct sockaddr_in6 *addr = (struct sockaddr_in6 *)sa;
+        res = inet_ntop(
+            addr->sin6_family,
+            (void*)&addr->sin6_addr,
+            buff, len);
+    } else {
+        struct sockaddr_in *addr = (struct sockaddr_in *)sa;
+        res = inet_ntop(
+            addr->sin_family, &addr->sin_addr.s_addr,
+            buff, len);
+    }
+#else
+    struct sockaddr_in *addr = (struct sockaddr_in *)sa;
+    res = inet_ntoa(addr->sin_addr);
+    erts_snprintf(buff,len,"%s", res);
+    res = buff;
+#endif
+    errno = myerrno;
+    return res;
+}
+
 void run(EpmdVars *g)
 {
   struct EPMD_SOCKADDR_IN iserv_addr[MAX_LISTEN_SOCKETS];
   int listensock[MAX_LISTEN_SOCKETS];
-  int num_sockets;
+#if defined(EPMD6)
+  char socknamebuf[INET6_ADDRSTRLEN];
+#else
+  char socknamebuf[INET_ADDRSTRLEN];
+#endif
+  int num_sockets = 0;
+  int nonfatal_sockets = 0;
   int i;
   int opt;
   unsigned short sport = g->port;
+  int bound = 0;
 
   node_init(g);
   g->conn = conn_init(g);
@@ -251,70 +294,96 @@ void run(EpmdVars *g)
   if (g->addresses != NULL && /* String contains non-separator characters if: */
       g->addresses[strspn(g->addresses," ,")] != '\000')
     {
-      char *tmp;
-      char *token;
-      int loopback_ok = 0;
+      char *tmp = NULL;
+      char *token = NULL;
 
-      if ((tmp = (char *)malloc(strlen(g->addresses) + 1)) == NULL)
+      /* Always listen on the loopback. */
+      SET_ADDR(iserv_addr[num_sockets],htonl(INADDR_LOOPBACK),sport);
+      num_sockets++;
+#if defined(EPMD6)
+      SET_ADDR6(iserv_addr[num_sockets],in6addr_loopback,sport);
+      num_sockets++;
+#endif
+
+      /* IPv4 and/or IPv6 loopback may not be present, for example if
+       * the protocol stack is explicitly disabled in the kernel.  If
+       * any sockets to this point fail, log the error but do not exit
+       * with failure.  If any socket after this (explicitly
+       * configured via addresses) fails, then consider the error
+       * fatal. */
+      nonfatal_sockets = num_sockets;
+
+	  if ((tmp = strdup(g->addresses)) == NULL)
 	{
 	  dbg_perror(g,"cannot allocate memory");
 	  epmd_cleanup_exit(g,1);
 	}
-      strcpy(tmp,g->addresses);
 
-      for(token = strtok(tmp,", "), num_sockets = 0;
+      for(token = strtok(tmp,", ");
 	  token != NULL;
-	  token = strtok(NULL,", "), num_sockets++)
+	  token = strtok(NULL,", "))
 	{
-	  struct EPMD_IN_ADDR addr;
-#ifdef HAVE_INET_PTON
-	  int ret;
+	  struct in_addr addr;
+#if defined(EPMD6)
+	  struct in6_addr addr6;
+	  struct sockaddr_storage *sa = &iserv_addr[num_sockets];
 
-	  if ((ret = inet_pton(FAMILY,token,&addr)) == -1)
+	  if (inet_pton(AF_INET6,token,&addr6) == 1)
 	    {
-	      dbg_perror(g,"cannot convert IP address to network format");
-	      epmd_cleanup_exit(g,1);
+	      SET_ADDR6(iserv_addr[num_sockets],addr6,sport);
 	    }
-	  else if (ret == 0)
-#elif !defined(EPMD6)
-	  if ((addr.EPMD_S_ADDR = inet_addr(token)) == INADDR_NONE)
+	  else if (inet_pton(AF_INET,token,&addr) == 1)
+	    {
+	      SET_ADDR(iserv_addr[num_sockets],addr.s_addr,sport);
+	    }
+	  else
+#else
+	  if ((addr.s_addr = inet_addr(token)) != INADDR_NONE)
+	    {
+	      SET_ADDR(iserv_addr[num_sockets],addr.s_addr,sport);
+	    }
+	  else
 #endif
 	    {
 	      dbg_tty_printf(g,0,"cannot parse IP address \"%s\"",token);
 	      epmd_cleanup_exit(g,1);
 	    }
 
-	  if (IS_ADDR_LOOPBACK(addr))
-	    loopback_ok = 1;
+#if defined(EPMD6)
+	  if (sa->ss_family == AF_INET6 && IN6_IS_ADDR_LOOPBACK(&addr6))
+	      continue;
 
-	  if (num_sockets - loopback_ok == MAX_LISTEN_SOCKETS - 1)
+	  if (sa->ss_family == AF_INET)
+#endif
+	  if (IS_ADDR_LOOPBACK(addr))
+	    continue;
+
+	  num_sockets++;
+
+	  if (num_sockets >= MAX_LISTEN_SOCKETS)
 	    {
 	      dbg_tty_printf(g,0,"cannot listen on more than %d IP addresses",
 			     MAX_LISTEN_SOCKETS);
 	      epmd_cleanup_exit(g,1);
 	    }
-
-	  SET_ADDR(iserv_addr[num_sockets],addr.EPMD_S_ADDR,sport);
 	}
 
       free(tmp);
-
-      if (!loopback_ok)
-	{
-	  SET_ADDR(iserv_addr[num_sockets],EPMD_ADDR_LOOPBACK,sport);
-	  num_sockets++;
-	}
     }
   else
     {
-      SET_ADDR(iserv_addr[0],EPMD_ADDR_ANY,sport);
-      num_sockets = 1;
+      SET_ADDR(iserv_addr[num_sockets],htonl(INADDR_ANY),sport);
+      num_sockets++;
+#if defined(EPMD6)
+      SET_ADDR6(iserv_addr[num_sockets],in6addr_any,sport);
+      num_sockets++;
+#endif
     }
 #ifdef HAVE_SYSTEMD_DAEMON
     }
 #endif /* HAVE_SYSTEMD_DAEMON */
 
-#if !defined(__WIN32__) && !defined(__OSE__)
+#if !defined(__WIN32__)
   /* We ignore the SIGPIPE signal that is raised when we call write
      twice on a socket closed by the other end. */
   signal(SIGPIPE, SIG_IGN);
@@ -339,13 +408,38 @@ void run(EpmdVars *g)
 #endif /* HAVE_SYSTEMD_DAEMON */
   for (i = 0; i < num_sockets; i++)
     {
-      if ((listensock[i] = socket(FAMILY,SOCK_STREAM,0)) < 0)
+      struct sockaddr *sa = (struct sockaddr *)&iserv_addr[i];
+#if defined(EPMD6)
+      size_t salen = (sa->sa_family == AF_INET6 ?
+              sizeof(struct sockaddr_in6) :
+              sizeof(struct sockaddr_in));
+#else
+      size_t salen = sizeof(struct sockaddr_in);
+#endif
+
+      if ((listensock[i] = socket(sa->sa_family,SOCK_STREAM,0)) < 0)
 	{
-	  dbg_perror(g,"error opening stream socket");
+	  switch (errno) {
+	      case EAFNOSUPPORT:
+	      case EPROTONOSUPPORT:
+	          continue;
+	      default:
+	          dbg_perror(g,"error opening stream socket");
+	          epmd_cleanup_exit(g,1);
+	  }
+	}
+
+#if HAVE_DECL_IPV6_V6ONLY
+      opt = 1;
+      if (sa->sa_family == AF_INET6 &&
+          setsockopt(listensock[i],IPPROTO_IPV6,IPV6_V6ONLY,&opt,
+              sizeof(opt)) <0)
+	{
+	  dbg_perror(g,"can't set IPv6 only socket option");
 	  epmd_cleanup_exit(g,1);
 	}
-      g->listenfd[i] = listensock[i];
-  
+#endif
+
       /*
        * Note that we must not enable the SO_REUSEADDR on Windows,
        * because addresses will be reused even if they are still in use.
@@ -368,42 +462,61 @@ void run(EpmdVars *g)
 	 in accept() waiting for the next request. */
 #if (defined(__WIN32__) || defined(NO_FCNTL))
       opt = 1;
-      /* Gives warning in VxWorks */
       if (ioctl(listensock[i], FIONBIO, &opt) != 0)
 #else
       opt = fcntl(listensock[i], F_GETFL, 0);
       if (fcntl(listensock[i], F_SETFL, opt | O_NONBLOCK) == -1)
-#endif /* __WIN32__ || VXWORKS */
-	dbg_perror(g,"failed to set non-blocking mode of listening socket %d",
-		   listensock[i]);
+#endif /* __WIN32__ */
+	dbg_perror(g,"failed to set non-blocking mode of listening socket %d on ipaddr %s",
+		   listensock[i], epmd_ntop(&iserv_addr[i],
+                                            socknamebuf, sizeof(socknamebuf)));
 
-      if (bind(listensock[i], (struct sockaddr*) &iserv_addr[i],
-	  sizeof(iserv_addr[i])) < 0)
+      if (bind(listensock[i], sa, salen) < 0)
 	{
 	  if (errno == EADDRINUSE)
 	    {
-	      dbg_tty_printf(g,1,"there is already a epmd running at port %d",
-			     g->port);
+	      dbg_tty_printf(g,1,"there is already a epmd running at port %d on ipaddr %s",
+			     g->port, epmd_ntop(&iserv_addr[i],
+                                                socknamebuf, sizeof(socknamebuf)));
 	      epmd_cleanup_exit(g,0);
 	    }
 	  else
 	    {
-	      dbg_perror(g,"failed to bind socket");
-	      epmd_cleanup_exit(g,1);
+              dbg_perror(g,"failed to bind on ipaddr %s",
+                         epmd_ntop(&iserv_addr[i],
+                                   socknamebuf, sizeof(socknamebuf)));
+              if (i >= nonfatal_sockets)
+                  epmd_cleanup_exit(g,1);
+              else
+              {
+                  close(listensock[i]);
+                  continue;
+              }
+
 	    }
 	}
 
       if(listen(listensock[i], SOMAXCONN) < 0) {
-          dbg_perror(g,"failed to listen on socket");
+          dbg_perror(g,"failed to listen on ipaddr %s",
+                     epmd_ntop(&iserv_addr[i],
+                               socknamebuf, sizeof(socknamebuf)));
           epmd_cleanup_exit(g,1);
       }
+      g->listenfd[bound++] = listensock[i];
       select_fd_set(g, listensock[i]);
     }
+  if (bound == 0) {
+      dbg_perror(g,"unable to bind any address");
+      epmd_cleanup_exit(g,1);
+  }
+  num_sockets = bound;
 #ifdef HAVE_SYSTEMD_DAEMON
     }
-    sd_notifyf(0, "READY=1\n"
-                  "STATUS=Processing port mapping requests...\n"
-                  "MAINPID=%lu", (unsigned long) getpid());
+    if (g->is_systemd) {
+      sd_notifyf(0, "READY=1\n"
+                    "STATUS=Processing port mapping requests...\n"
+                    "MAINPID=%lu", (unsigned long) getpid());
+    }
 #endif /* HAVE_SYSTEMD_DAEMON */
 
   dbg_tty_printf(g,2,"entering the main select() loop");
@@ -443,12 +556,12 @@ void run(EpmdVars *g)
 	}
 
 	for (i = 0; i < num_sockets; i++)
-	  if (FD_ISSET(listensock[i],&read_mask)) {
-	    if (do_accept(g, listensock[i]) && g->active_conn < g->max_conn) {
+	  if (FD_ISSET(g->listenfd[i],&read_mask)) {
+	    if (do_accept(g, g->listenfd[i]) && g->active_conn < g->max_conn) {
 	      /*
 	       * The accept() succeeded, and we have at least one file
 	       * descriptor still free, which means that another accept()
-	       * could succeed. Go do do another select(), in case there
+	       * could succeed. Go do another select(), in case there
 	       * are more incoming connections waiting to be accepted.
 	       */
 	      goto select_again;
@@ -477,7 +590,7 @@ void run(EpmdVars *g)
 
 /*
  *  This routine read as much of the packet as possible and
- *  if completed calls "do_request()" to fullfill the request.
+ *  if completed calls "do_request()" to fulfill the request.
  *
  */
 
@@ -617,14 +730,24 @@ static int do_accept(EpmdVars *g,int listensock)
     return conn_open(g,msgsock);
 }
 
+static void bump_creation(Node* node)
+{
+    if (++node->cr_counter < 4)
+        node->cr_counter = 4;
+}
+static unsigned int get_creation(Node* node)
+{
+    if (node->highvsn >= 6) {
+        return node->cr_counter;  /* 4..(2^32-1)*/
+    }
+    else {
+        return node->cr_counter % 3 + 1;   /* 1..3 */
+    }
+}
+
 /* buf is actually one byte larger than bsize,
    giving place for null termination */
-static void do_request(g, fd, s, buf, bsize)
-     EpmdVars *g;
-     int fd;
-     Connection *s;
-     char *buf;
-     int bsize;
+static void do_request(EpmdVars *g, int fd, Connection *s, char *buf, int bsize)
 {
   char wbuf[OUTBUF_SIZE];	/* Buffer for writing */
   int i;
@@ -658,8 +781,10 @@ static void do_request(g, fd, s, buf, bsize)
 	unsigned char protocol;
 	unsigned short highvsn;
 	unsigned short lowvsn;
+        unsigned int creation;
 	int namelen;
 	int extralen;
+        int replylen;
 	char *name; 
 	char *extra;
 	eport = get_int16(&buf[1]);
@@ -689,21 +814,24 @@ static void do_request(g, fd, s, buf, bsize)
 
 	extra = &buf[11+namelen+2];
 	extra[extralen]='\000';
-	wbuf[0] = EPMD_ALIVE2_RESP;
-	if ((node = node_reg2(g, namelen, name, fd, eport, nodetype, protocol,
-			      highvsn, lowvsn, extralen, extra)) == NULL) {
-	    wbuf[1] = 1; /* error */
-	    put_int16(99, wbuf+2);
-	} else {
-	    wbuf[1] = 0; /* ok */
-	    put_int16(node->creation, wbuf+2);
-	}
+        node = node_reg2(g, namelen, name, fd, eport, nodetype, protocol,
+                         highvsn, lowvsn, extralen, extra);
+        creation = node ? get_creation(node) :  99;
+        wbuf[1] = node ? 0 : 1; /* ok | error */
+        if (highvsn >= 6) {
+            wbuf[0] = EPMD_ALIVE2_X_RESP;
+            put_int32(creation, wbuf+2);
+            replylen = 6;
+        }
+        else {
+            wbuf[0] = EPMD_ALIVE2_RESP;
+            put_int16(creation, wbuf+2);
+            replylen = 4;
+        }
   
-	if (g->delay_write)		/* Test of busy server */
-	  sleep(g->delay_write);
-
-	if (reply(g, fd, wbuf, 4) != 4)
+	if (!reply(g, fd, wbuf, replylen))
 	  {
+            node_unreg(g, name);
 	    dbg_tty_printf(g,1,"** failed to send ALIVE2_RESP for \"%s\"",
 			   name);
 	    return;
@@ -761,7 +889,7 @@ static void do_request(g, fd, s, buf, bsize)
 		offset += 2;
 		memcpy(wbuf + offset,node->extra,node->extralen);
 		offset += node->extralen;
-		if (reply(g, fd, wbuf, offset) != offset)
+		if (!reply(g, fd, wbuf, offset))
 		  {
 		    dbg_tty_printf(g,1,"** failed to send PORT2_RESP (ok) for \"%s\"",name);
 		    return;
@@ -771,7 +899,7 @@ static void do_request(g, fd, s, buf, bsize)
 	    }
 	}
 	wbuf[1] = 1; /* error */
-	if (reply(g, fd, wbuf, 2) != 2)
+	if (!reply(g, fd, wbuf, 2))
 	  {
 	    dbg_tty_printf(g,1,"** failed to send PORT2_RESP (error) for \"%s\"",name);
 	    return;
@@ -789,7 +917,7 @@ static void do_request(g, fd, s, buf, bsize)
 	i = htonl(g->port);
 	memcpy(wbuf,&i,4);
 
-	if (reply(g, fd,wbuf,4) != 4)
+	if (!reply(g, fd,wbuf,4))
 	  {
 	    dbg_tty_printf(g,1,"failed to send NAMES_RESP");
 	    return;
@@ -810,7 +938,7 @@ static void do_request(g, fd, s, buf, bsize)
 	    if (r < 0)
 		goto failed_names_resp;
 	    len += r;
-	    if (reply(g, fd, wbuf, len) != len)
+	    if (!reply(g, fd, wbuf, len))
 	      {
 	      failed_names_resp:
 		dbg_tty_printf(g,1,"failed to send NAMES_RESP");
@@ -832,7 +960,7 @@ static void do_request(g, fd, s, buf, bsize)
 
 	i = htonl(g->port);
 	memcpy(wbuf,&i,4);
-	if (reply(g, fd,wbuf,4) != 4)
+	if (!reply(g, fd,wbuf,4))
 	  {
 	    dbg_tty_printf(g,1,"failed to send DUMP_RESP");
 	    return;
@@ -853,7 +981,7 @@ static void do_request(g, fd, s, buf, bsize)
 	      if (r < 0)
 		  goto failed_dump_resp;
 	      len += r + 1;
-	      if (reply(g, fd,wbuf,len) != len)
+	      if (!reply(g, fd,wbuf,len))
 	      {
 	      failed_dump_resp:
 		dbg_tty_printf(g,1,"failed to send DUMP_RESP");
@@ -876,7 +1004,7 @@ static void do_request(g, fd, s, buf, bsize)
 	      if (r < 0)
 		  goto failed_dump_resp2;
 	      len += r + 1;
-	      if (reply(g, fd,wbuf,len) != len)
+	      if (!reply(g, fd,wbuf,len))
 	      {
 	      failed_dump_resp2:
 		dbg_tty_printf(g,1,"failed to send DUMP_RESP");
@@ -896,12 +1024,12 @@ static void do_request(g, fd, s, buf, bsize)
 
       if (!g->brutal_kill && (g->nodes.reg != NULL)) {
 	  dbg_printf(g,0,"Disallowed KILL_REQ, live nodes");
-	  if (reply(g, fd,"NO",2) != 2)
+	  if (!reply(g, fd,"NO",2))
 	      dbg_printf(g,0,"failed to send reply to KILL_REQ");
 	  return;
       }
 
-      if (reply(g, fd,"OK",2) != 2)
+      if (!reply(g, fd,"OK",2))
 	dbg_printf(g,0,"failed to send reply to KILL_REQ");
       dbg_tty_printf(g,1,"epmd killed");
       conn_close_fd(g,fd);	/* We never return to caller so close here */
@@ -931,18 +1059,20 @@ static void do_request(g, fd, s, buf, bsize)
 
 	if ((node_fd = node_unreg(g,name)) < 0)
 	  {
-	    if (reply(g, fd,"NOEXIST",7) != 7)
+	    if (!reply(g, fd,"NOEXIST",7))
 	      {
 		dbg_tty_printf(g,1,"failed to send STOP_RESP NOEXIST");
 		return;
 	      }
 	    dbg_tty_printf(g,1,"** sent STOP_RESP NOEXIST");
 	  }
+        else
+          {
+            conn_close_fd(g,node_fd);
+            dbg_tty_printf(g,1,"epmd connection stopped");
+          }
 
-	conn_close_fd(g,node_fd);
-	dbg_tty_printf(g,1,"epmd connection stopped");
-
-	if (reply(g, fd,"STOPPED",7) != 7)
+	if (!reply(g, fd,"STOPPED",7))
 	  {
 	    dbg_tty_printf(g,1,"failed to send STOP_RESP STOPPED");
 	    return;
@@ -987,18 +1117,9 @@ static int conn_open(EpmdVars *g,int fd)
   int i;
   Connection *s;
 
-#ifdef VXWORKS
-  /*
-   * Since file descriptors are global on VxWorks, we might get an fd that
-   * does not fit in the FD_SET.
-   *
-   * Note: This test would be harmless on Unix, but would fail on Windows
-   * because socket are numbered differently and FD_SETs are implemented
-   * differently.
-   */
+#if !defined(__WIN32__)
   if (fd >= FD_SETSIZE) {
-      dbg_tty_printf(g,0,"file descriptor %d: too high for FD_SETSIZE=%d",
-		     fd,FD_SETSIZE);
+      dbg_tty_printf(g,0,"fd does not fit in fd_set fd=%d, FD_SETSIZE=%d",fd, FD_SETSIZE);
       close(fd);
       return EPMD_FALSE;
   }
@@ -1006,15 +1127,6 @@ static int conn_open(EpmdVars *g,int fd)
 
   for (i = 0; i < g->max_conn; i++) {
     if (g->conn[i].open == EPMD_FALSE) {
-      struct sockaddr_in si;
-      struct sockaddr_in di;
-#ifdef HAVE_SOCKLEN_T
-      socklen_t st;
-#else
-      int st;
-#endif
-      st = sizeof(si);
-
       g->active_conn++;
       s = &g->conn[i];
      
@@ -1025,20 +1137,7 @@ static int conn_open(EpmdVars *g,int fd)
       s->open = EPMD_TRUE;
       s->keep = EPMD_FALSE;
 
-      /* Determine if connection is from localhost */
-      if (getpeername(s->fd,(struct sockaddr*) &si,&st) ||
-	  st < sizeof(si)) {
-	  /* Failure to get peername is regarded as non local host */
-	  s->local_peer = EPMD_FALSE;
-      } else {
-	  /* Only 127.x.x.x and connections from the host's IP address
-	     allowed, no false positives */
-	  s->local_peer =
-	      (((((unsigned) ntohl(si.sin_addr.s_addr)) & 0xFF000000U) ==
-	       0x7F000000U) ||
-	       (getsockname(s->fd,(struct sockaddr*) &di,&st) ?
-	       EPMD_FALSE : si.sin_addr.s_addr == di.sin_addr.s_addr));
-      }
+      s->local_peer = conn_local_peer_check(g, s->fd);
       dbg_tty_printf(g,2,(s->local_peer) ? "Local peer connected" :
 		     "Non-local peer connected");
 
@@ -1046,7 +1145,7 @@ static int conn_open(EpmdVars *g,int fd)
       s->got  = 0;
       s->mod_time = current_time(g); /* Note activity */
 
-      s->buf = (char *)malloc(INBUF_SIZE);
+      s->buf = malloc(INBUF_SIZE);
 
       if (s->buf == NULL) {
 	dbg_printf(g,0,"epmd: Insufficient memory");
@@ -1064,6 +1163,64 @@ static int conn_open(EpmdVars *g,int fd)
   return EPMD_FALSE;
 }
 
+static int conn_local_peer_check(EpmdVars *g, int fd)
+{
+  struct EPMD_SOCKADDR_IN si;
+  struct EPMD_SOCKADDR_IN di;
+
+  struct sockaddr_in *si4 = (struct sockaddr_in *)&si;
+  struct sockaddr_in *di4 = (struct sockaddr_in *)&di;
+
+#if defined(EPMD6)
+  struct sockaddr_in6 *si6 = (struct sockaddr_in6 *)&si;
+  struct sockaddr_in6 *di6 = (struct sockaddr_in6 *)&di;
+#endif
+
+#ifdef HAVE_SOCKLEN_T
+  socklen_t st;
+#else
+  int st;
+#endif
+
+  st = sizeof(si);
+#ifdef __clang_analyzer__
+  /* CodeChecker does not seem to understand getpeername writes to 'si' */
+  memset(&si, 0, sizeof(si));
+#endif
+
+  /* Determine if connection is from localhost */
+  if (getpeername(fd,(struct sockaddr*) &si,&st) ||
+	  st > sizeof(si)) {
+	  /* Failure to get peername is regarded as non local host */
+	  return EPMD_FALSE;
+  }
+
+  /* Only 127.x.x.x and connections from the host's IP address
+	 allowed, no false positives */
+#if defined(EPMD6)
+  if (si.ss_family == AF_INET6 && IN6_IS_ADDR_LOOPBACK(&(si6->sin6_addr)))
+	  return EPMD_TRUE;
+
+  if (si.ss_family == AF_INET)
+#endif
+  if ((((unsigned) ntohl(si4->sin_addr.s_addr)) & 0xFF000000U) ==
+	  0x7F000000U)
+	  return EPMD_TRUE;
+
+  if (getsockname(fd,(struct sockaddr*) &di,&st))
+	  return EPMD_FALSE;
+
+#if defined(EPMD6)
+  if (si.ss_family == AF_INET6)
+      return IN6_ARE_ADDR_EQUAL( &(si6->sin6_addr), &(di6->sin6_addr));
+  if (si.ss_family == AF_INET)
+#endif
+  return si4->sin_addr.s_addr == di4->sin_addr.s_addr;
+#if defined(EPMD6)
+  return EPMD_FALSE;
+#endif
+}
+
 static int conn_close_fd(EpmdVars *g,int fd)
 {
   int i;
@@ -1122,8 +1279,8 @@ static int node_unreg(EpmdVars *g,char *name)
   for (; node; prev = &node->next, node = node->next)
     if (is_same_str(node->symname, name))
       {
-	dbg_tty_printf(g,1,"unregistering '%s:%d', port %d",
-		       node->symname, node->creation, node->port);
+	dbg_tty_printf(g,1,"unregistering '%s:%u', port %d",
+		       node->symname, get_creation(node), node->port);
 
 	*prev = node->next;	/* Link out from "reg" list */
 
@@ -1157,8 +1314,8 @@ static int node_unreg_sock(EpmdVars *g,int fd)
   for (; node; prev = &node->next, node = node->next)
     if (node->fd == fd)
       {
-	dbg_tty_printf(g,1,"unregistering '%s:%d', port %d",
-		       node->symname, node->creation, node->port);
+	dbg_tty_printf(g,1,"unregistering '%s:%u', port %d",
+		       node->symname, get_creation(node), node->port);
 
 	*prev = node->next;	/* Link out from "reg" list */
 
@@ -1186,19 +1343,8 @@ static int node_unreg_sock(EpmdVars *g,int fd)
 }
 
 /*
- *  Finding a node slot and a (name,creation) name is a bit tricky.
- *  We try in order
- *
- *  1. If the name was used before and we can reuse that slot but use
- *     a new "creation" digit in the range 1..3.
- *
- *  2. We try to find a new unused slot.
- *
- *  3. We try to use an used slot this isn't used any longer.
- *     FIXME: The criteria for *what* slot to steal should be improved.
- *     Perhaps use the oldest or something.
+ * Register a new node
  */
-
 static Node *node_reg2(EpmdVars *g,
 		       int namelen,
 		       char* name,
@@ -1268,7 +1414,7 @@ static Node *node_reg2(EpmdVars *g,
       }
 
   /* Try to find the name in the used queue so that we
-     can change "creation" number 1..3 */
+     can change "creation" number */
 
   prev = NULL;
 
@@ -1297,9 +1443,8 @@ static Node *node_reg2(EpmdVars *g,
 
 	g->nodes.unreg_count--;
 
-	/* When reusing we change the "creation" number 1..3 */
-
-	node->creation = node->creation % 3 + 1;
+	/* When reusing we change the "creation" number */
+	bump_creation(node);
 
 	break;
       }
@@ -1314,6 +1459,7 @@ static Node *node_reg2(EpmdVars *g,
 	  (g->debug && (g->nodes.unreg_count > DEBUG_MAX_UNREG_COUNT)))
 	{
 	  /* MAX_UNREG_COUNT > 1 so no need to check unreg_tail */
+          ASSERT(g->nodes.unreg != NULL);
 	  node = g->nodes.unreg;	/* Take first == oldest */
 	  g->nodes.unreg = node->next; /* Link out */
 	  g->nodes.unreg_count--;
@@ -1326,7 +1472,8 @@ static Node *node_reg2(EpmdVars *g,
 	      exit(1);
 	    }
 
-	  node->creation = (current_time(g) % 3) + 1; /* "random" 1-3 */
+	  node->cr_counter = current_time(g); /* "random" */
+          bump_creation(node);
 	}
     }
 
@@ -1345,11 +1492,11 @@ static Node *node_reg2(EpmdVars *g,
   select_fd_set(g, fd);
 
   if (highvsn == 0) {
-    dbg_tty_printf(g,1,"registering '%s:%d', port %d",
-		   node->symname, node->creation, node->port);
+    dbg_tty_printf(g,1,"registering '%s:%u', port %d",
+		   node->symname, get_creation(node), node->port);
   } else {
-    dbg_tty_printf(g,1,"registering '%s:%d', port %d",
-		   node->symname, node->creation, node->port);
+    dbg_tty_printf(g,1,"registering '%s:%u', port %d",
+		   node->symname, get_creation(node), node->port);
     dbg_tty_printf(g,1,"type %d proto %d highvsn %d lowvsn %d",
 		   nodetype, protocol, highvsn, lowvsn);
   }      
@@ -1370,7 +1517,9 @@ static time_t current_time(EpmdVars *g)
 
 static int reply(EpmdVars *g,int fd,char *buf,int len)
 {
-  int val;
+  char* p = buf;
+  int nbytes = len;
+  int val, ret;
 
   if (len < 0)
     {
@@ -1381,15 +1530,27 @@ static int reply(EpmdVars *g,int fd,char *buf,int len)
   if (g->delay_write)		/* Test of busy server */
     sleep(g->delay_write);
 
-  val = write(fd,buf,len);
-  if (val < 0)
-    dbg_perror(g,"error in write");
-  else if (val != len)
-    dbg_printf(g,0,"could only send %d bytes out of %d to fd %d",val,len,fd);
+  for (;;) {
+      val = write(fd, p, nbytes);
+      if (val == nbytes) {
+          ret = 1;
+          break;
+      }
+      if (val < 0) {
+          if (errno == EINTR)
+              continue;
+          dbg_perror(g,"error in write, errno=%d", errno);
+          ret = 0;
+          break;
+      }
+      dbg_printf(g,0,"could only send %d bytes out of %d to fd %d",val,nbytes,fd);
+      p += val;
+      nbytes -= val;
+  }
 
   dbg_print_buf(g,buf,len);
 
-  return val;
+  return ret;
 }
       
 
@@ -1469,8 +1630,8 @@ static void print_names(EpmdVars *g)
 
   for (node = g->nodes.reg; node; node = node->next)
     {
-      fprintf(stderr,"*****     active name     \"%s#%d\" at port %d, fd = %d\r\n",
-	      node->symname, node->creation, node->port, node->fd);
+      fprintf(stderr,"*****     active name     \"%s#%u\" at port %d, fd = %d\r\n",
+	      node->symname, get_creation(node), node->port, node->fd);
       count ++;
     }
 
@@ -1480,8 +1641,8 @@ static void print_names(EpmdVars *g)
 
   for (node = g->nodes.unreg; node; node = node->next)
     {
-      fprintf(stderr,"*****     old/unused name \"%s#%d\"\r\n",
-	      node->symname, node->creation);
+      fprintf(stderr,"*****     old/unused name \"%s#%u\"\r\n",
+	      node->symname, get_creation(node));
       count ++;
     }
 
