@@ -1,7 +1,9 @@
 %%
 %% %CopyrightBegin%
 %%
-%% Copyright Ericsson AB 1996-2024. All Rights Reserved.
+%% SPDX-License-Identifier: Apache-2.0
+%%
+%% Copyright Ericsson AB 1996-2025. All Rights Reserved.
 %%
 %% Licensed under the Apache License, Version 2.0 (the "License");
 %% you may not use this file except in compliance with the License.
@@ -48,8 +50,9 @@
 
 %% fence code
 -export([single_line_fence_code_test/1, multiple_line_fence_code_test/1,
+         single_line_fence_code_no_language_test/1, single_line_fence_code_no_language_spaces_test/1,
          paragraph_between_fence_code_test/1, fence_code_ignores_link_format_test/1,
-         fence_code_with_spaces/1]).
+         fence_code_with_spaces/1, fence_code_with_tabs/1]).
 
 %% br
 -export([start_with_br_test/1, multiple_br_followed_by_paragraph_test/1,
@@ -68,7 +71,7 @@
          list_format_with_italics_in_sentence/1, list_format_with_bold_in_sentence/1,
          new_lines_test/1, format_separator_test/1, list_with_format/1, multi_word_format_test/1,
          multiline_link/1, multiline_link_not_allowed/1, inline_mfa_link/1,
-         escaped_character/1]).
+         escaped_character/1, parens_with_italics/1]).
 
 %% Bullet lists
 -export([singleton_bullet_list/1, singleton_bullet_list_followed_new_paragraph/1, singleton_bullet_list_with_format/1,
@@ -188,9 +191,11 @@ code_tests() ->
 fence_code_tests() ->
   [single_line_fence_code_test,
    multiple_line_fence_code_test,
+   single_line_fence_code_no_language_test,
+   single_line_fence_code_no_language_spaces_test,
    paragraph_between_fence_code_test,
    fence_code_ignores_link_format_test,
-   fence_code_with_spaces
+   fence_code_with_spaces, fence_code_with_tabs
   ].
 
 br_tests() ->
@@ -231,7 +236,8 @@ format_tests() ->
       multiline_link,
       multiline_link_not_allowed,
       inline_mfa_link,
-      escaped_character
+      escaped_character,
+      parens_with_italics
     ].
 
 bullet_list_tests() ->
@@ -492,7 +498,7 @@ single_line_fence_code_test(_Conf) ->
 ```erlang
 test() -> ok.
 ```",
-    Result = [ code(~"test() -> ok.\n")],
+    Result = [ code(~"test() -> ok.\n", [{class, ~"language-erlang"}])],
     compile_and_compare(Input, Result).
 
 multiple_line_fence_code_test(_Conf) ->
@@ -501,9 +507,24 @@ multiple_line_fence_code_test(_Conf) ->
 test() ->
   ok.
 ```",
-    Result = [ code(~"test() ->\n  ok.\n")],
+    Result = [ code(~"test() ->\n  ok.\n", [{class, ~"language-erlang"}])],
     compile_and_compare(Input, Result).
 
+single_line_fence_code_no_language_test(_Conf) ->
+    Input = ~"
+```
+test() -> ok.
+```",
+    Result = [ code(~"test() -> ok.\n")],
+    compile_and_compare(Input, Result).
+
+single_line_fence_code_no_language_spaces_test(_Conf) ->
+    Input = ~"
+```\s\s
+test() -> ok.
+```",
+    Result = [ code(~"test() -> ok.\n")],
+    compile_and_compare(Input, Result).
 
 paragraph_between_fence_code_test(_Conf) ->
     Input = ~"This is a test:
@@ -512,7 +533,7 @@ test() ->
   ok.
 ```",
     Result = [p(~"This is a test:"),
-                         code(~"test() ->\n  ok.\n")],
+                         code(~"test() ->\n  ok.\n", [{class, ~"language-erlang"}])],
     compile_and_compare(Input, Result).
 
 fence_code_ignores_link_format_test(_Conf) ->
@@ -521,15 +542,23 @@ fence_code_ignores_link_format_test(_Conf) ->
 [foo](bar)
 ```",
     Result = [p(~"This is a test:"),
-              code(~"[foo](bar)\n")],
+              code(~"[foo](bar)\n", [{class, ~"language-erlang"}])],
     compile_and_compare(Input, Result).
 
 fence_code_with_spaces(_Config) ->
     Input =
-~"  ```erlang
+~"  ```erlang\s\s
   [foo](bar)
 ```",
-    Result = [code(~"  [foo](bar)\n")],
+    Result = [code(~"  [foo](bar)\n", [{class, ~"language-erlang"}])],
+    compile_and_compare(Input, Result).
+
+fence_code_with_tabs(_Config) ->
+    Input =
+~"  ```erlang\ttrailing
+  [foo](bar)
+```",
+    Result = [code(~"  [foo](bar)\n", [{class, ~"language-erlang"}])],
     compile_and_compare(Input, Result).
 
 start_with_br_test(_Conf) ->
@@ -728,6 +757,25 @@ escaped_character(_Config) ->
               inline_code(~"`test`")]),
   compile_and_compare(Input, Result).
 
+parens_with_italics(_Config) ->
+  Input = ~"Test ( _Optional_). ( _Optional_ .) ( _Optional_ ). (_Optional_). (_Optional._)",
+  Result = p([~"Test ( ", it(~"Optional"), ~"). ( "
+                        , it(~"Optional"), ~" .) ( "
+                        , it(~"Optional"), ~" ). ("
+                        , it(~"Optional"), ~"). ("
+                        , it(~"Optional."), ~")"]),
+  compile_and_compare(Input, Result),
+
+  Input1 = ~"Test ( __Optional__). ( __Optional__ .) ",
+  Input2 = ~"( __Optional__ ). (__Optional__). (__Optional.__)",
+  InputEm = <<Input1/binary, Input2/binary>>,
+  ResultEm = p([~"Test ( ", em(~"Optional"), ~"). ( "
+                          , em(~"Optional"), ~" .) ( "
+                          , em(~"Optional"), ~" ). ("
+                          , em(~"Optional"), ~"). ("
+                          , em(~"Optional."), ~")"]),
+  compile_and_compare(InputEm, ResultEm).
+
 list_with_format(_Config) ->
   Input = ~"- **`--help` (or `-h`)** - Print this message and exit.",
   Result = ul([li(p([em([inline_code(~"--help"),
@@ -1019,10 +1067,14 @@ header(Level, Text) when is_integer(Level) ->
     {HeadingLevelAtom, [], [Text]}.
 
 code(X) ->
-    {pre,[],[inline_code(X)]}.
+    code(X, []).
+code(X, Attrs) when is_list(X) ->
+    {pre,[],[{code,Attrs,X}]};
+code(X, Attrs) ->
+    {pre,[],[{code,Attrs,[X]}]}.
 
 table(Table) when is_list(Table) ->
-    {pre,[], [inline_code(Table)]}.
+    {pre,[], [{code, [{class, ~"table"}], Table}]}.
 
 inline_code(X) when is_list(X) ->
     {code,[],X};
